#================================
# BVH v1.1
# Copyright Reevan Mckay 2000-2001
# All Rights Reserved
# reevan@blender.nl
#
# IMPORTANT:
#	You probably need the FULL python installation to run this script.
#		If you have problems, get the python installation from
#		http://www.python.org
#
# PURPOSE:
#	To allow the import of Biovision Hierarchical motion capture data (.bvh).
#	Primarily for use in GameBlender, but also useful for regular animation.
#
# USAGE:
#	[1]	Run the script once and press the HIERARCHY button to determine the hierarchical
#		structure of the character.  It will be printed out in a file named HIERARCHY.HTML
#		which can be read in a standard html browser.
#	[2] Create a series of empties with the same names as the joints in the output file.
#	[3]	Link the objects together in the hierarchy described in the HIERARCHY.HTML file,
#		using SHIFT+CONTROL+PKEY (NOT CONTROL+PKEY).  When you do this, you will notice
#		that the child snaps to the location of the parent.  You will probably want to
#		 drag them apart again to make it easier to visualize the skeleton structure.
#	[4]	Insert a ROT key for all empties
#	[5]	Insert a LOCROT key for the root (usually called HIPS or PELVIS)
#	[6]	Run the script and press the BONE button.  This will automatically adjust the base
#		position of the empties, rename their IPO curves and clear their rotations.
#	[7]	Press the MOVE button.  This will load the motion data and apply it to the IPO curves
#		of the skeleton.
#	[8]	It is not necessary to press the BONE button again if the next motion file you use
#		has the same hierarchy information (e.g. two files from the same mocap session, or
#		imported from the same object from a different animation package such as A:M or Maya
#
# LIMITATIONS:
#	* This reader doesn't do anything with the "End Site" information yet.  This means you
#	will have to manually pose the rest position for the head, hands and feet.
#	* The BVH format assumes Y to be "up".  Not Z, as is used in Blender.  This means that
#	motions will be imported with incorrect orientation.  A temporary fix is to parent the root of
#	the character to an empty and rotate that empty by 90 degrees around the x axis.
#	* The hierarchy you design in Blender must have the same stucture and joint names as
#	the one in the file you wish to import.  Workarounds are being investigated.
#	* Motion import will fail if the joint empties do not have the right ipo curves in them.
#	* The motion capture data will not work with motion blur enabled, or if you change the
#	time scale (Old/New buttons in the AnimButtons window).
#	* It's not particularly fault-tolerant.  If there is a problem with importing, the
#	script is unlikely to warn you.  There might be some information in the blender console
#	window however.
#
# NOTES:
#	A sample BVH file is included.  To save it from the text editor, use SHIFT+ALT+FKEY>>SAVE AS
#	I'm trying to keep the source for this relatively unified.
#	If you would like to contribute, or make modifications, send your mods
#		to me and I will incorporate them into the next release for everyone to
#		enjoy.
#	Please don't distribute modified copies of this file without contacting me
#		first!
#
# HISTORY:
#	Aug 24, 01: Fixed default settings and removed Animation:Master specific code
#	Oct 29, 00: First release for internal NaN use
#
#================================

import Blender
from Blender.Draw import *
from Blender.BGL import *
import string
import math

DEG_TO_RAD=			math.pi/180.0
RAD_TO_DEG=			180.0/math.pi

EVENT_EXIT=			1
EVENT_START=		2
EVENT_END=			3
EVENT_LIBPATH=		4
EVENT_ACTNAME=		5
EVENT_REFRESH=		6
EVENT_SAVE=			7
EVENT_LOAD=			8
EVENT_MOVE=			9
EVENT_BONE=			10
EVENT_HIER=			11

ERROR_NONE=			0
ERROR_UNKNOWN=		1
ERROR_EOF=			2
ERROR_EXPECT_OB=	3
ERROR_EXPECT_CB=	4
ERROR_EXPECT_IDENTIFIER=	5
ERROR_EXPECT_ASSIGN=6

TOKEN_OPENBRACE=	"T_OPEN"
TOKEN_CLOSEBRACE=	"T_CLOSE"
TOKEN_NUMBER=		"T_NUMBER"
TOKEN_IDENTIFIER=	"T_IDENTIFIER"
TOKEN_UNKNOWN=		"T_UNKNOWN"
TOKEN_EOF=			"T_EOF"
TOKEN_OFFSET=		"T_OFFSET"
TOKEN_ROOT=			"T_ROOT"
TOKEN_HIERARCHY=	"T_HEIRARCHY"
TOKEN_MOTION=		"T_MOTION"
TOKEN_CHANNELS=		"T_CHANNELS"
TOKEN_JOINT=		"T_JOINT"
TOKEN_FRAMES=		"T_FRAMES"
TOKEN_FRAME=		"T_FRAME"
TOKEN_TIME=			"T_TIME"
TOKEN_COLON=		"T_COLON"
TOKEN_END=			"T_END"
TOKEN_SITE=			"T_SITE"
TOKEN_XPOS=			"T_XPOS"
TOKEN_YPOS=			"T_YPOS"
TOKEN_ZPOS=			"T_ZPOS"
TOKEN_XROT=			"T_XROT"
TOKEN_YROT=			"T_YROT"
TOKEN_ZROT=			"T_ZROT"
TOKEN_CHANNEL=		"T_CHANNEL"

MESSAGE_NONE=		0
MESSAGE_OPEN=		1
MESSAGE_CLOSE=		2
MESSAGE_SINGLE=		3


VALID_KEYWORDS=		[
						[TOKEN_HIERARCHY,"HIERARCHY"],
						[TOKEN_MOTION,"MOTION"],
						[TOKEN_ROOT,"ROOT"],
						[TOKEN_OFFSET,"OFFSET"],
						[TOKEN_CHANNELS,"CHANNELS"],
						[TOKEN_JOINT,"JOINT"],
						[TOKEN_FRAMES,"Frames"],
						[TOKEN_FRAME,"Frame"],
						[TOKEN_TIME,"Time"],
						[TOKEN_END,"End"],
						[TOKEN_SITE,"Site"],
						[TOKEN_CHANNEL,"Xposition"],
						[TOKEN_CHANNEL,"Yposition"],
						[TOKEN_CHANNEL,"Zposition"],
						[TOKEN_CHANNEL,"Xrotation"],
						[TOKEN_CHANNEL,"Yrotation"],
						[TOKEN_CHANNEL,"Zrotation"],

					]

VALID_TOKENS=		[	[TOKEN_OPENBRACE,"{"],
						[TOKEN_CLOSEBRACE,"}"],
						[TOKEN_COLON,":"],				
					]
					
VALID_DIGITS=		"+-.0123456789"
VALID_FULLDIGITS=	"+-.0123456789e"
VALID_CHARS=		"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"
VALID_FULLCHARS=	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789"
VALID_WHITESPACE=	"\m\n\r\t "

SCALING	=	1

g_Prefix=Create ("")
g_FrameMax	=	100
g_LibPath = Create("ballerina.bvh")
g_ActName = Create("NewAction")
g_ActionMenu = Create (0)
g_StartFrame=Create (1)
g_EndFrame=Create (30)
g_Scaling=Create (1)

g_FileOpen = 0

g_Actions=[]

#	Count Actions in File/Make Action List
#	Delete Action from File
#	Save Action to File



#=================================
def RVShutdown ():
#=================================
	Blender.Draw.Exit()
	print ("Shutting Down...\n")


#=================================
def RVDrawInterface ():
#=================================
	global g_FrameMax
	global g_LibPath
	global g_ActName
	global g_StartFrame
	global g_EndFrame
	global g_ActionMenu
	global g_Scaling
	global g_Prefix

	glClearColor (0.6,0.6,0.6,0)
#	glColor3ub (255,255,255)
	glClear (GL_COLOR_BUFFER_BIT)

#	Draw LibFile Button
	g_LibPath=String ("Lib File: ",EVENT_REFRESH,32,156,228,16,g_LibPath.val,255,"The location to save the library file")

	try:
		theFile=open(g_LibPath.val,"ra")
		glColor3ub (255,255,255)
		glRasterPos2i (280,160)
		Text ("File Found")
		theFile.close()
	except IOError:
		glColor3ub (0,0,0)
		glRasterPos2i (280,160)
		Text ("File Not Found")

	g_Prefix=String ("Prefix: ",EVENT_REFRESH,100,104,128,16,g_Prefix.val,255,"Prefix for objects in hierarchy")
	g_Scaling=Number ("Scaling: ",EVENT_REFRESH,100,84,128,16,g_Scaling.val,-1000.00,1000.00)
	Button ("Move",EVENT_MOVE,100,64,64,16)
	Button ("Bone",EVENT_BONE,164,64,64,16)
	Button ("Hierarchy",EVENT_HIER,100,44,64,16)

#	Draw Exit Button
	Button ("Exit",EVENT_EXIT,164,44,64,16)

#	Draw Action Menu
#	g_ActionMenu = Menu ("No Actions Loaded%t",EVENT_REFRESH,180,44,128,16,g_ActionMenu.val,"Available Actions")


#=================================
def RVHandleEvent (eventNum, eventMod):
#=================================
	pass

#=================================
def RVHandleButton (eventNum):
#=================================

	global g_ActName
	global g_StartFrame
	global g_EndFrame

	if (eventNum==EVENT_EXIT):
		RVShutdown()
	if (eventNum==EVENT_MOVE):
		RVLoadFile(EVENT_MOVE)
	if (eventNum==EVENT_BONE):
		RVLoadFile(EVENT_BONE)
	if (eventNum==EVENT_HIER):
		RVLoadFile(EVENT_HIER)
	if (eventNum==EVENT_START):
		Draw ()
	if (eventNum==EVENT_END):
		Draw()
	if (eventNum==EVENT_SAVE):
		RVSaveFile (RVMakeAction(g_ActName.val,(g_EndFrame.val-g_StartFrame.val)))
		Draw()
	if (eventNum==EVENT_REFRESH):
		Draw()

#=================================
def	RVStartup():
#=================================
	print ("Starting Up...")
	Register (RVDrawInterface,None,RVHandleButton)

#=================================
def	RVMakeBone (name,start,end):
#=================================
	pass

	# Find start-end vector
	vector[0]=end[0]-start[0]
	vector[1]=end[1]-start[1]
	vector[2]=end[2]-start[2]

	# Make Mesh
	
	mesh=NMesh.GetRaw (name)
	vertList=[]

	vertList.append (NMesh.Vert(0,0,0))
	vertList.append (NMesh.Vert(vector[0],vector[1],vector[2]))	
	mesh.verts=vertList

	faceList=[]
	face=NMesh.Face()
	face.v.append(mesh.verts[0])
	face.v.append(mesh.verts[1])
	faceList.append(face)
	mesh.faces=faceList

	NMesh.PutRaw (mesh,name,1)
	

# 

#=================================
def	RVGetCurve (ipo,name):
#=================================

	index=0
	seachName="RotX"

	if type (ipo)==None:
		print "IPO ERROR!"
		return

	if name=="Xrotation":
		searchName="RotX"
	elif name=="Yrotation":
		searchName="RotY"
	elif name=="Zrotation":
		searchName="RotZ"
	elif name=="Xposition":
		searchName="LocX"
	elif name=="Yposition":
		searchName="LocY"
	elif name=="Zposition":
		searchName="LocZ"
	else:
		searchName=name
	

	for c in ipo.curves:
	#	print (c.name)
		if c.name==searchName:
			return index
		index=index+1
		
	return -1


#=================================
def	RVApplyOffsets (inData):
#=================================

	global g_Scaling

	for j in inData:
		curOb=Blender.Object.Get(j[0])
		index=0
		if (type(curOb)!=type(None)):
			for i in curOb.ipo.curves:
				pointList=[]
				newPoint=Blender.Ipo.BezTriple()
				newPoint.h1[0]=0
				newPoint.h1[1]=0

				newPoint.h2[0]=0
				newPoint.h2[1]=0

				newPoint.pt[0]=0
				newPoint.pt[1]=0

				newPoint.h1t="Auto"
				newPoint.h2t="Auto"
				pointList.append(newPoint)
				curOb.ipo.curves[index].points=pointList
				index=index+1
	
				curOb.rot[0]=0
				curOb.rot[1]=0
				curOb.rot[2]=0

				curOb.loc[0]=j[1][0]*g_Scaling.val
				curOb.loc[1]=j[1][1]*g_Scaling.val
				curOb.loc[2]=j[1][2]*g_Scaling.val


#=================================
def RVApplyMotion (inData):
#=================================
#	return

	global g_Scaling
	curIpo=None
	pointCount=0

	for j in inData:
		print ("Getting IPO: "+j[0])
		curOb=Blender.Object.Get(j[0])
		if type(curOb)==type(None):
			continue
		curOb.ipo.name=j[0]
		curIpo=curOb.ipo
		if type(curIpo)==type(None):
			continue
		pointCount=0
		hasx=hasy=hasz=0
		for c in j[2]:
			print ("Setting CUR: "+c[0])
			curCurve=curIpo.curves[RVGetCurve (curIpo,c[0])]
			if type(curCurve)==type(None):
				continue
			curX=0
			pointList=[]
			pointCount=0
			if c[0]=="Xrotation":
				hasx=1
			if c[0]=="Yrotation":
				hasy=1
			if c[0]=="Zrotation":
				hasz=1
		
			for s in c[1]:
				pointCount=pointCount+1
				curX=curX+1
				newPoint=Blender.Ipo.BezTriple()
				newPoint.h1[0]=0
				newPoint.h1[1]=0

				newPoint.h2[0]=0
				newPoint.h2[1]=0
				newPoint.pt[0]=curX

				if (c[0]=="Xposition" or c[0]=="Yposition" or c[0]=="Zposition"):
					newPoint.pt[1]=s*g_Scaling.val
				else:			
					newPoint.pt[1]=s

				newPoint.h1t="Auto"
				newPoint.h2t="Auto"
				pointList.append(newPoint)
			curCurve.points=pointList


		
		if hasx==1 and hasy==1 and hasz==1:

			xrot=curIpo.curves[RVGetCurve(curIpo,"RotX")].points
			yrot=curIpo.curves[RVGetCurve(curIpo,"RotY")].points
			zrot=curIpo.curves[RVGetCurve(curIpo,"RotZ")].points
			for i in range(pointCount):
				eul=[xrot[i].pt[1],yrot[i].pt[1],zrot[i].pt[1]]

				for jj in range(3):
					while eul[jj]<0:
						eul[jj]=eul[jj]+360.00
					while eul[jj]>=360:
						eul[jj]=eul[jj]-360.00
	
					
				eul[0]=eul[0]*DEG_TO_RAD
				eul[1]=eul[1]*DEG_TO_RAD
				eul[2]=eul[2]*DEG_TO_RAD



				xmat=RVAxisAngleToMat3([1,0,0,eul[0]])
				ymat=RVAxisAngleToMat3([0,1,0,eul[1]])
				zmat=RVAxisAngleToMat3([0,0,1,eul[2]])


				mat=[[1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,1.0]]

#				Standard BVH multiplication order
				mat=RVMatMult3 (zmat,mat)
				mat=RVMatMult3 (xmat,mat)
				mat=RVMatMult3 (ymat,mat)

#				Screwy Animation Master BVH multiplcation order
#				mat=RVMatMult3 (ymat,mat)
#				mat=RVMatMult3 (xmat,mat)
#				mat=RVMatMult3 (zmat,mat)

				eul=RVMat3ToXYZ (mat)	
	
				eul[0]=eul[0]*RAD_TO_DEG
				eul[1]=eul[1]*RAD_TO_DEG
				eul[2]=eul[2]*RAD_TO_DEG


				xrot[i].pt[1]=-eul[0]/10
				yrot[i].pt[1]=-eul[1]/10
				zrot[i].pt[1]=-eul[2]/10


			curIpo.curves[RVGetCurve(curIpo,"RotX")].points=xrot
			curIpo.curves[RVGetCurve(curIpo,"RotY")].points=yrot
			curIpo.curves[RVGetCurve(curIpo,"RotZ")].points=zrot


				

#=================================
def	RVReadBVH(theFile,action):
#=================================
	
	bvhData=[]
	#	Name
	#	FrameCount
	#	FrameTime
	#	Object List

	while (1):
		token=RVGetNextToken (theFile)
		if token[0]==TOKEN_EOF:
			break
		elif token[0]==TOKEN_HIERARCHY:
			bvhData=RVReadHierarchy (theFile,bvhData)
		elif token[0]==TOKEN_MOTION:
			bvhData=RVReadMotion (theFile,bvhData)
		else:
			print ("Expected MOTION or HIERARCHY. "+token[1]+" ["+token[0]+"] found instead.")	
			break

	if (action==EVENT_HIER):	
		
		outFile=open("Hierarchy.html","wb")
		for j in bvhData:
			outFile.write (j[0]+"<BR>\n")
			outFile.write ("<BLOCKQUOTE>\n")
			for c in j[2]:
				outFile.write ("\t"+c[0]+"<BR>\n")
			outFile.write ("</BLOCKQUOTE>\n")
#				for s in c[1]:
#					outFile.write (str(s)+"<BR>\n")
		outFile.close()


	if (action==EVENT_BONE):
		RVApplyOffsets (bvhData)
	if (action==EVENT_MOVE):
		RVApplyMotion (bvhData)
		RVAMWorkaround (bvhData)


#=================================
def RVAMWorkaround (bvhData):
#	Fixes a fucked up problem with
#	Animation:Master 8.5 mocap export
#=================================

	return

	print "Working around!"
	for j in bvhData:
		hasLoc=0
		hasRot=0
		print "Working around "+j[0]
		for c in j[2]:
			if c[0]=="Xposition":
				hasLoc=1
			if c[0]=="Zrotation":
				hasRot=1
		if hasLoc==1 and hasRot==1:
			print ("Has Loc and Rot")
			# get the curve and points
			curIpo=Blender.Object.Get(j[0]).ipo
			xrot=curIpo.curves[RVGetCurve(curIpo,"RotX")].points
			yrot=curIpo.curves[RVGetCurve(curIpo,"RotY")].points
			zrot=curIpo.curves[RVGetCurve(curIpo,"RotZ")].points
			xloc=curIpo.curves[RVGetCurve(curIpo,"LocX")].points
			yloc=curIpo.curves[RVGetCurve(curIpo,"LocY")].points
			zloc=curIpo.curves[RVGetCurve(curIpo,"LocZ")].points

			pointCount=0
			for s in c[1]:
				pointCount=pointCount+1

			print "Fixing "+str(pointCount)+"points"
			for i in range (pointCount):
				xpt=xloc[i].pt[1]
				ypt=yloc[i].pt[1]
				zpt=zloc[i].pt[1]

				xloc[i].pt[1]=xpt
				yloc[i].pt[1]=ypt
				zloc[i].pt[1]=zpt

			curIpo.curves[RVGetCurve(curIpo,"LocX")].points=xloc
			curIpo.curves[RVGetCurve(curIpo,"LocY")].points=yloc
			curIpo.curves[RVGetCurve(curIpo,"LocZ")].points=zloc

#=================================
def	RVReadOffset (theFile):
#	Assumes word OFFSET has been read
#=================================
	offset=[0.0,0.0,0.0]
	for v in range(3): 
		token=RVGetNextToken (theFile)
		if token[0]!=TOKEN_NUMBER:
			print ("Expected a value. "+token[1]+" ["+token[0]+"] found instead.")	
			return ERROR_UNKNOWN
		offset[v]=string.atof (token[1])
	return offset

#=================================
def	RVReadHNode (theFile,root,inNode,parentOffset):
#=================================

	global g_Prefix

	ob=["Base",[0.0,0.0,0.0],[],[],"JOINT"]
	# Name, Offset, Channels, Channel Data,type


	#---------------
	# If this is a root, read the root identifier
	#---------------

	if root==1:
		ob[4]="ROOT"
		token=RVGetNextToken(theFile)
		if token[0]!=TOKEN_ROOT:
			print ("Expected \"ROOT\". "+token[1]+" ["+token[0]+"] found instead.")	
			return

	#---------------
	# Find the object's identifier
	#---------------

	token=RVGetNextToken (theFile)
	if token[0]!=TOKEN_IDENTIFIER:
		print ("Expected identifier. "+token[1]+" ["+token[0]+"] found instead.")	
		return
	ob[0]=g_Prefix.val+token[1]

	#---------------
	# Read an open brace
	#---------------

	if RVVerifyNextToken (theFile,TOKEN_OPENBRACE)!=ERROR_NONE:
		return 
		
	#---------------
	# Read Offset
	#---------------

	if RVVerifyNextToken (theFile,TOKEN_OFFSET)!=ERROR_NONE:
		return 
	
	ob[1]=RVReadOffset(theFile)
	ob[1][0]=ob[1][0]+parentOffset[0]
	ob[1][1]=ob[1][1]+parentOffset[1]
	ob[1][2]=ob[1][2]+parentOffset[2]

	#---------------
	# Read Channels
	#---------------

	if RVVerifyNextToken (theFile,TOKEN_CHANNELS)!=ERROR_NONE:
		return

	token=RVGetNextToken (theFile)
	if token[0]!=TOKEN_NUMBER:
		print ("Expected an integer. "+token[1]+" ["+token[0]+"] found instead.")	
		return
	for i in range(string.atoi(token[1])):
		newChannel=["",[]]
		token=RVGetNextToken (theFile)
		if token[0]!=TOKEN_CHANNEL:
			print ("Expected a channel identifier. "+token[1]+" ["+token[0]+"] found instead.")	
			return
		newChannel[0]=token[1]
		ob[2].append(newChannel)
	
	inNode.append (ob)

	#---------------
	# Read JOINTS or Close Brace
	#---------------
	while (1):
		token=RVGetNextToken (theFile)
		if token[0]==TOKEN_JOINT:
			RVReadHNode (theFile,0,inNode,ob[1])
		elif token[0]==TOKEN_CLOSEBRACE:
			return
		elif token[0]==TOKEN_END:
			RVReadEndSite (theFile,ob[1])			
		else:
			print ("Expected \"JOINT\" or \"}\". "+token[1]+" ["+token[0]+"] found instead.")	
			return 

	return ob

#=================================
def	RVReadEndSite (theFile,parentOffset):
#	Assumes the word "END" has been read
#=================================

	if (RVVerifyNextToken (theFile,TOKEN_SITE)!=ERROR_NONE):
		return ERROR_UNKNOWN
	
	if (RVVerifyNextToken (theFile,TOKEN_OPENBRACE)!=ERROR_NONE):
		return ERROR_UNKNOWN
	 
	while 1:
		token=RVGetNextToken(theFile)
		if token[0]==TOKEN_OFFSET:
			RVReadOffset (theFile)
		elif token[0]==TOKEN_CLOSEBRACE:
			return ERROR_NONE
		else:
			print ("Expected \"OFFSET\" or \"}\". "+token[1]+" ["+token[0]+"] found instead.")	
			return ERROR_UNKNOWN
			
	return ERROR_NONE

#=================================
def	RVVerifyNextToken (theFile,tokenType):
#=================================
	token=RVGetNextToken(theFile)
	if token[0]!=tokenType:
		print ("Expected token type ["+tokenType+"].  Found ["+token[0]+"] instead.")		
		return ERROR_UNKNOWN
	else:
		return ERROR_NONE
#=================================
def	RVReadHierarchy(theFile,bvhData):
#=================================
	

	RVReadHNode (theFile,1,bvhData,[0.0,0.0,0.0])
	return bvhData

#=================================
def	RVEulToMat3(eul):
#=================================

#	print ("euler in: ",eul)
	mat=[[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]

	x=eul[0]
	y=eul[1]
	z=eul[2]

	mat[0][0]=math.cos(y)*math.cos(z)
	mat[0][1]=math.cos(y)*math.sin(z)
	mat[0][2]=-math.sin(y)

	mat[1][0]=math.sin(x)*math.sin(y)*math.cos(z)-math.cos(x)*math.sin(z)
	mat[1][1]=math.sin(x)*math.sin(y)*math.sin(z)+math.cos(x)*math.cos(z)
	mat[1][2]=math.sin(x)*math.cos(y)

	mat[2][0]=math.cos(x)*math.sin(y)*math.cos(z)+math.sin(x)*math.sin(z)
	mat[2][1]=math.cos(x)*math.sin(y)*math.sin(z)-math.sin(x)*math.sin(z)
	mat[2][2]=math.cos(x)*math.cos(y)

	return mat



#=================================
def RVMat3ToXYZ (mat):
#=================================
	PI=3.14159

	thetaY=math.asin (mat[0][2])
	if (thetaY<PI/2):
		if (thetaY>-PI/2):
			thetaX=math.atan2(-mat[1][2],mat[2][2])
			thetaZ=math.atan2(-mat[0][1],mat[0][0])
		else:
			thetaX=-math.atan2(mat[1][0],mat[1][1])
			thetaZ=0.0
	else:
		thetaX=math.atan2(mat[1][0],mat[1][1])
		thetaZ=0.0
	return [thetaX,thetaY,thetaZ]


#=================================
def	RVMat3ToXZY (mat):
#=================================
	PI=3.14159

	thetaZ=math.asin (-mat[0][1])
	if (thetaZ<PI/2):
		if (thetaZ>-PI/2):
			thetaX=math.atan2(mat[2][1],mat[1][1])
			thetaY=math.atan2(mat[0][2],mat[0][0])
		else:
			thetaX=-math.atan2(-mat[2][0],mat[2][2])
			thetaY=0.0
	else:
		thetaX=math.atan2(-mat[2][0],mat[2][2])
		thetaY=0.0

	return [thetaX,thetaY,thetaZ]

#=================================
def RVMat3ToYXZ (mat):
#=================================
	PI=3.14159

	thetaX=math.asin (-mat[1][2])
	if (thetaX<PI/2):
		if (thetaX>-PI/2):
			thetaY=math.atan2(mat[0][2],mat[2][2])
			thetaZ=math.atan2(mat[1][0],mat[1][1])
		else:
			thetaY=-math.atan2(-mat[0][1],mat[0][0])
			thetaZ=0.0
	else:
		thetaY=math.atan2(-mat[0][1],mat[0][0])
		thetaZ=0.0
	return [thetaX,thetaY,thetaZ]

#=================================
def RVMat3ToYZX (mat):
#=================================
	PI=3.14159

	thetaZ=math.asin (mat[1][0])
	if (thetaZ<PI/2):
		if (thetaZ>-PI/2):
			thetaY=math.atan2(-mat[2][0],mat[0][0])
			thetaX=math.atan2(-mat[1][2],mat[1][1])
		else:
			thetaY=-math.atan2(mat[2][1],mat[2][2])
			thetaX=0.0
	else:
		thetaY=math.atan2(mat[2][1],mat[2][2])
		thetaX=0.0
	return [thetaX,thetaY,thetaZ]


#=================================
def RVMat3ToZXY (mat):
#=================================
	PI=3.14159

	thetaX=math.asin (mat[2][1])
	if (thetaX<PI/2):
		if (thetaX>-PI/2):
			thetaZ=math.atan2(-mat[0][1],mat[1][1])
			thetaY=math.atan2(-mat[2][0],mat[2][2])
		else:
			thetaZ=-math.atan2(mat[0][2],mat[0][0])
			thetaY=0.0
	else:
		thetaZ=math.atan2(mat[0][2],mat[0][0])
		thetaY=0.0

	return [thetaX,thetaY,thetaZ]

#=================================
def RVMat3ToZYX (mat):
#=================================
	PI=3.14159

	thetaY=math.asin (-mat[2][0])
	if (thetaY<PI/2):
		if (thetaY>-PI/2):
			thetaZ=math.atan2(mat[1][0],mat[0][0])
			thetaX=math.atan2(mat[2][1],mat[2][2])
		else:
			thetaZ=-math.atan2(mat[0][1],mat[0][2])
			thetaX=0.0
	else:
		thetaZ=math.atan2(-mat[0][1],mat[0][2])
		thetaX=0.0

	return [thetaX,thetaY,thetaZ]

#=================================
def	RVMat3ToEul (mat):
#	Note: This doesn't work.  Not
#	even a little bit.
#=================================
	FLT_EPSILON=0.0001

	sin1=-mat[2][0]
	cos1=math.sqrt(1.0-sin1*sin1)
	
	if math.fabs(cos1)>FLT_EPSILON:
		sin2=mat[2][1]/cos1
		cos2=mat[2][2]/cos1
		sin3=mat[1][0]/cos1
		cos3=mat[0][0]/cos1
	else:
		sin2=-mat[1][2]
		cos2=tmat[1][1]
		sin3=0.0
		cos3=1.0

	eul=[	math.atan2(sin3,cos3),
			math.atan2(sin1,cos1),
			math.atan2(sin2,cos2)]

	return eul

#=================================
def	RVReadMotion (theFile,inData):
#=================================
	if RVVerifyNextToken(theFile,TOKEN_FRAMES)!=ERROR_NONE:
		return ERROR_UNKNOWN
	if RVVerifyNextToken(theFile,TOKEN_COLON)!=ERROR_NONE:
		return ERROR_UNKNOWN

	# Read Number of frames
	token=RVGetNextToken(theFile)
	if token[0]!=TOKEN_NUMBER:
		print ("Expected integer.  Found ["+token[0]+"] instead.")		
		return ERROR_UNKNOWN
		
	frameCount=string.atoi (token[1])

	if RVVerifyNextToken(theFile,TOKEN_FRAME)!=ERROR_NONE:
		return ERROR_UNKNOWN
	if RVVerifyNextToken(theFile,TOKEN_TIME)!=ERROR_NONE:
		return ERROR_UNKNOWN
	if RVVerifyNextToken(theFile,TOKEN_COLON)!=ERROR_NONE:
		return ERROR_UNKNOWN

	# Read frame time
	token=RVGetNextToken(theFile)
	if token[0]!=TOKEN_NUMBER:
		print ("Expected float.  Found ["+token[0]+"] instead.")		
		return ERROR_UNKNOWN
	
	sampleCount=0
	for j in inData:
		for s in j[2]:
			sampleCount=sampleCount+1

	print "Found "+str(sampleCount)+" samples"

	for f in range(frameCount):
		for j in inData:
			for s in j[2]:
				# Read value
				token=RVGetNextToken(theFile)
				if token[0]!=TOKEN_NUMBER:
					print ("Expected integer.  Found ["+token[0]+"] instead.")		
					return
				# Append to list
			#	print inData[inData.index(j)][2][j[2].index(s)][1]
				inData[inData.index(j)][2][j[2].index(s)][1].append (string.atof(token[1]))
			#	j[1].append(string.atof(token[1]))

	# Read key block
	frameTime=string.atof (token[1])
	return inData

#=================================
def	RVLoadFile(event):
#=================================
	try:
		theFile=open (g_LibPath.val,"ra")
	except IOError:
		return None
	
	RVReadBVH (theFile,event)
	theFile.close
	

#=================================
def	RVSkipWhitespace (theFile):
#	skip 	\n
#			\t
#			" "
#			\r
#=================================

	while (1):
		try:
			char=theFile.read(1)
		except IOError:
			return ERROR_UNKNOWN

		if char=="":
			return ERROR_EOF
		if string.find (VALID_WHITESPACE,char)==-1:
			theFile.seek (-1,1)
			return ERROR_NONE
 

#=================================
def	RVGetIdentifier (theFile,firstByte):
#=================================
	
	token = [TOKEN_IDENTIFIER,firstByte]
	
	while (1):
		curByte=theFile.read(1)
		if curByte=="":
			return token	
		if string.find (VALID_FULLCHARS,curByte)==-1:
			theFile.seek (-1,1)
			break
		token[1]=token[1]+curByte

	# Check if this is a keyword
	for k in VALID_KEYWORDS:
		if k[1]==token[1]:
			return k

	return token

#=================================
def	RVGetValue (theFile,firstByte):
#
#=================================
	token = [TOKEN_NUMBER,firstByte]
	
	while (1):
		curByte=theFile.read(1)
		if curByte=="":
			return token	
		if string.find (VALID_FULLDIGITS,curByte)==-1:
			theFile.seek (-1,1)
			break
		token[1]=token[1]+curByte
	return token


#=================================
def RVGetNextToken1 (theFile):
#=================================
	result=RVGetNextToken1(theFile)
	print ("TOKEN: "+result[1])
	return result

#=================================
def	RVGetNextToken (theFile):
#=================================
	token = [TOKEN_UNKNOWN,"__UNKNOWN"]
	curString=""
	foundToken=-1

	if RVSkipWhitespace (theFile)==ERROR_EOF:
		return [TOKEN_EOF,"__EOF"]
	
	curByte=theFile.read(1)
	if curByte=="":	
		return [TOKEN_EOF,"__EOF"]


	#	If character
	if string.find (VALID_CHARS,curByte)!=-1:
		return RVGetIdentifier (theFile,curByte)

	#	If valChar
	if string.find (VALID_DIGITS,curByte)!=-1:
		return RVGetValue (theFile,curByte)

	#	Look for a reserved token
	while (1):
		foundToken=0
		for t in VALID_TOKENS:
			if t[1]==curString+curByte:
				foundToken=1
				token=t
				curString=curString+curByte
				break


	#	If we're not building a valid token,
	#	return the last valid one we found
		if foundToken==0:
			theFile.seek (-1,1)
			return token

		curByte=theFile.read(1)
		if curByte=="":
			return [TOKEN_EOF,"__EOF"]			

	return token

#=================================
def RVMatMult3 (mat1,mat2):
#=================================

	mat3=[[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]

	for i in range(3):
		for k in range(3):
			for j in range(3):
				mat3[i][k]=mat3[i][k]+mat1[i][j]*mat2[j][k]

	return mat3

#=================================
def	RVAxisAngleToMat3 (rot4):
#	Takes a direction vector and
#	a rotation (in rads) and
#	returns the rotation matrix.
#	Graphics Gems I p. 466:
#=================================

	mat3=[[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]

	if math.fabs(rot4[3])>0.01:
		s=math.sin(rot4[3])
		c=math.cos(rot4[3])
		t=1.0-math.cos(rot4[3])
	else:
		s=rot4[3]
		c=1.0
		t=0.0

	x=rot4[0]
	y=rot4[1]
	z=rot4[2]

	mat3[0][0]=t*x*x+c
	mat3[0][1]=t*x*y+s*z
	mat3[0][2]=t*x*z-s*y

	mat3[1][0]=t*x*y-s*z
	mat3[1][1]=t*y*y+c
	mat3[1][2]=t*y*z+s*x

	mat3[2][0]=t*x*z+s*y
	mat3[2][1]=t*y*z-s*x
	mat3[2][2]=t*z*z+c
	
	return mat3
	
#=================================
# main ()
#=================================
RVStartup()